package fernandoVisitorSet;

import joeq.Compiler.Quad.*;
import joeq.Compiler.Quad.SSA.*;
import joeq.Compiler.Dataflow.*;
import joeq.Main.Helper;
import joeq.Class.*;
import joeq.Util.Templates.*;

public class CFGSculptor {
    ControlFlowGraph cfg = null;

    public CFGSculptor(ControlFlowGraph cfg) {
        this.cfg = cfg;
    }

    public ControlFlowGraph getCFG() {
        return this.cfg;
    }

    /**
     * Splits edges in two, by inserting a new basic block between the old 
     * ones. Some analyses and transformatioons are simpler if there is
     * never a control-flow edge that leads from a node with multiple
     * successors to a node with multiple predecessors. The objective of
     * this method is to remove such edges.
     */
    public void splitEdges() {
        java.util.LinkedList newBlockList = new java.util.LinkedList();
        ListIterator.BasicBlock bbIterator = this.cfg.postOrderOnReverseGraphIterator();
        while(bbIterator.hasNext()) {
            BasicBlock bb = bbIterator.nextBasicBlock();
            if(bb.getNumberOfPredecessors() > 1) {
                List.BasicBlock predList = bb.getPredecessors();
                ListIterator.BasicBlock preds = predList.basicBlockIterator();
                while(preds.hasNext()) {
                    BasicBlock predBlock = (BasicBlock)preds.next();
                    if(predBlock.getNumberOfSuccessors() > 1) {
                        BasicBlock newBlock = this.cfg.createBasicBlock(1, 1, 0, null);
                        newBlock.addPredecessor(predBlock);
                        newBlock.addSuccessor(bb);
                        newBlockList.add(newBlock);
                    }
                }
            }
        }
        // updating of the control flow graph:
        while(newBlockList.size() > 0) {
            BasicBlock predBlock, succBlock;
            BasicBlock newBlock = (BasicBlock)newBlockList.removeFirst();

            List.BasicBlock preds = newBlock.getPredecessors();
            predBlock = preds.getBasicBlock(0);
            predBlock.addSuccessor(newBlock);

            List.BasicBlock succs = newBlock.getSuccessors();
            succBlock = succs.getBasicBlock(0);
            succBlock.addPredecessor(newBlock);

            predBlock.removeSuccessor(succBlock);
            succBlock.removePredecessor(predBlock);
        }
    }


    /**
     * Add the quad to the end of the basic block. If the last instruction
     * of the basic block is a branch, the quad will be inserted before
     * it.
     * @param bb the basic block that will receive the quad.
     * @param quad the quad to be inserted.
     */
    public void insertBeforeBranch(BasicBlock bb, Quad quad) {
         // these operations avoid inserting move instructions in the entry block
        if(bb.isEntry()) {
            java.util.LinkedList newBlockList = new java.util.LinkedList();
            BasicBlock newBlock = this.cfg.createBasicBlock(1, 1, 0, null);
            newBlock.addPredecessor(bb);
            List.BasicBlock succList = bb.getSuccessors();
            ListIterator.BasicBlock succs = succList.basicBlockIterator();
            while(succs.hasNext()) {
                BasicBlock succBB = succs.nextBasicBlock();
                succBB.addPredecessor(newBlock);
                newBlock.addSuccessor(succBB);
                newBlockList.add(succBB);
            }
            while(newBlockList.size() > 0) {
                BasicBlock block = (BasicBlock)newBlockList.removeFirst();
                bb.removeSuccessor(block);
                block.removePredecessor(bb);
            }
            bb.addSuccessor(newBlock);
            bb = newBlock;
        }

        Quad lastQuad = bb.getLastQuad();
        if(lastQuad != null) {
            Operator op = lastQuad.getOperator();
            if(op instanceof Operator.Branch)
                bb.addQuad(bb.getQuadIndex(lastQuad), quad);
            else
                bb.appendQuad(quad);
        }
        else
            bb.appendQuad(quad);
    }


    /**
     * This method converts the encapsulated graph to SSA representation.
     * In this representation, every variable is defined just once.
     */
    public void convertoToSSA() {
        EnterSSA.markSSARegisterFlags(this.cfg);
        EnterSSA essa = new EnterSSA();
        essa.visitCFG(this.cfg);
    }


    /**
     * This method adapts the control flow graph to generate the interference
     * graph. First it converts the cfg to the ssa form. After, it converts
     * it back to a sequence of instructions without phi functions. This last
     * step is accomplished by substituting the phi functions by move
     * instructions, which are inserted in the end of the predecessor blocks
     * in which the phi functions are defined.
     */
    public void adaptCFG() {
        splitEdges();

        // SSA conversion:
        convertoToSSA();

        // remove phi functions and insert move instructions:
        ListIterator.BasicBlock bbIterator = this.cfg.postOrderOnReverseGraphIterator();
        while(bbIterator.hasNext()) {
            BasicBlock bb = bbIterator.nextBasicBlock();
            java.util.LinkedList phiQuads = new java.util.LinkedList();
            ListIterator.Quad qIter = bb.iterator();
            while(qIter.hasNext()) {
                Quad quad = qIter.nextQuad();
                Operator op = quad.getOperator();
                if(op instanceof Operator.Phi.PHI) {
                    Operator.Phi.PHI phi = (Operator.Phi.PHI)op;
                    Operand.RegisterOperand dest = Operator.Phi.PHI.getDest(quad);
                    UnmodifiableList.RegisterOperand uReg = phi.getUsedRegisters(quad);
                    Operand.BasicBlockTableOperand pred = Operator.Phi.PHI.getPreds(quad);
                    for(int i = 0; i < uReg.size(); i++) {
                        // get source register: src
                        Operand.RegisterOperand src = Operator.Phi.PHI.getSrc(quad, i);
                        // add move dest src to predBlock
                        Quad move = Operator.Move.create(this.cfg.getNewQuadID(), dest.getRegister(), src.getRegister(), src.getType());
                        // get the block that originates src: predBlock
                        BasicBlock predBlock = Operator.Phi.PHI.getPred(quad, i);
                        this.insertBeforeBranch(predBlock, move);

                        // remove the phi function
                        phiQuads.add(quad);
                    }
                }
            }
            while(phiQuads.size() > 0) {
                Quad quad = (Quad)phiQuads.removeFirst();
                bb.removeQuad(quad);
            }
        }
    }

}
