package fernandoVisitorSet;

import FernandoGraph.*;

import joeq.Compiler.Quad.*;
import joeq.Compiler.Quad.SSA.*;
import joeq.Compiler.Dataflow.*;
import joeq.Main.Helper;
import joeq.Class.*;
import joeq.Util.Templates.*;

import java.io.IOException;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.HashSet;

/**
 * This visitor builds interference graphs from control flow graphs.
 * Each interference will be output in a file with the same name of the
 * method that generated the file.
 */
public class InterferenceGraphBuilder implements ControlFlowGraphVisitor {

    private boolean toGraph = false;

    private int numRegisters = 0;

    private boolean isSSA = false;

    private Hashtable liveIn = new Hashtable();

    private RegisterAllocatorGraph graph;

    private String className = "";

    /**
     * Constructor method.
     * @param isSSA true if the target program must be converted to SSA form before
     * register allocation. False otherwise.
     * @param isMCS true if register allocation will be performed by coloring of
     * chordal graphs. False if iterated register coalescing should be used instead.
     * @param numRegisters maximum number of registers allowed in the allocation. If
     * this number is not sufficient, spill will be performed.
     * @param toGraph true if the output of the algorithm should be recorded in a
     * graphviz .dot file. False if only staticsts about the allocation must be
     * printed.
     * @param className the name of the class that is providing the methods to be
     * parsed.
     */
    public InterferenceGraphBuilder(boolean isSSA, boolean isMCS, int numRegisters, boolean toGraph, String className) {
        this.className = className;
        this.className.replace('.', '_');
        this.toGraph = toGraph;
        this.isSSA = isSSA;
        this.numRegisters = numRegisters;
        if(isMCS)
            graph = new MCSGraph();
        else
            graph = new IteratedCoalescing();
    }

    /**
     * Constructor method.
     * @param isSSA true if the target program must be converted to SSA form before
     * register allocation. False otherwise.
     * @param isMCS true if register allocation will be performed by coloring of
     * chordal graphs. False if iterated register coalescing should be used instead.
     * @param numRegisters maximum number of registers allowed in the allocation. If
     * this number is not sufficient, spill will be performed.
     * @param toGraph true if the output of the algorithm should be recorded in a
     * graphviz .dot file. False if only staticsts about the allocation must be
     * printed.
     */
    public InterferenceGraphBuilder(boolean isSSA, boolean isMCS, int numRegisters, boolean toGraph) {
        this.toGraph = toGraph;
        this.isSSA = isSSA;
        this.numRegisters = numRegisters;
        if(isMCS)
            graph = new MCSGraph();
        else
            graph = new IteratedCoalescing();
    }

    /**
     * This method resolves the interferences between basic blocks. An edge in
     * the interference graph will be generated for each pair of registers
     * alive at the end of the basic block.
     */
    public void buildInterBlockInterferences(ControlFlowGraph cfg) {
        // Solve liveness analysis:
        LivenessAnalysis la = new LivenessAnalysis();
        Solver solver = new IterativeSolver();
        jwutil.graphs.EdgeGraph eg = new jwutil.graphs.EdgeGraph(new jwutil.graphs.ReverseGraph(cfg, java.util.Collections.singleton(cfg.exit())));
        solver.initialize(la, eg);
        solver.solve();

        // compute live sets and interferences:
        ListIterator.BasicBlock bbIterator = cfg.postOrderOnReverseGraphIterator();
        while(bbIterator.hasNext()) {
            BasicBlock bb = bbIterator.nextBasicBlock();
            UnionBitVectorFact f = (UnionBitVectorFact)solver.getDataflowValue(bb);
            jwutil.math.BitString bs = f.getBitString();
            addInSet(bb, bs);
            for(int i = 0; i < (bs.length() - 1); i++)
                for(int j = i+1; j < bs.length(); j++)
                    if(bs.get(i) && bs.get(j)) {
                        String p1 = "R" + (i - 1);
                        String p2 = "R" + (j - 1);
                        graph.addEdge(p1, p2);
                    }
        }
    }


    /**
     * This method adds to the table <CODE>liveIn</CODE> the variables that are
     * alive at the begining of the given block. Such information is necessary in
     * order to compute the live interferences in the interior of basic blocks.
     * In order to compute the out set of a basic block, just do the union of the
     * in sets of the successors of that basic block.
     */
    private void addInSet(BasicBlock bb, jwutil.math.BitString bs) {
        HashSet liveSet = new HashSet();
        for(int i = 0; i < bs.length() - 1; i++)
            if(bs.get(i))
                liveSet.add("R" + (i - 1));
        this.liveIn.put(bb, liveSet);
    }


    /**
     * this method computes the set of variables that are alive at the end of
     * the given block.
     * The out set of a basic block is the union of the in set of its successors.
     * The in set of the successors have been computed in the method
     * <CODE>buildInterBlockInterferences</CODE>.
     */
    private HashSet computeOutSet(BasicBlock bb) {
        HashSet outSet = new HashSet();
        List.BasicBlock lbb = bb.getSuccessors();
        ListIterator.BasicBlock libb = lbb.basicBlockIterator();
        while(libb.hasNext()) {
            BasicBlock bbAux = libb.nextBasicBlock();
            HashSet inSet = (HashSet)this.liveIn.get(bbAux);
            Iterator iterator = inSet.iterator();
            while(iterator.hasNext())
                outSet.add(iterator.next());
        }
        return outSet;
    }


    /**
     * This method adds to the interference graph the interferences between registers
     * inside the basic block. It is linear in the number of quads of the basic block.
     * A backward analysis is performed. In this analysis the initial set of live
     * variables is the out set of the basic block.
     */
    private void buildInterferencesInBlock(BasicBlock bb) {
        // add to the list of living variables the variables that are in the out set
        // of the basic block
        java.util.HashSet liveVariables = computeOutSet(bb);

        ListIterator.Quad quadList = bb.backwardIterator();
        while(quadList.hasNext()) {
            Quad quad = (Quad)quadList.next();
            // marking the move instructions for coalescing:
            if(quad.getOperator() instanceof Operator.Move) {
                Operand srcOp = Operator.Move.getSrc(quad);
                if(srcOp instanceof Operand.RegisterOperand) {
                    String src = ((Operand.RegisterOperand)srcOp).getRegister().toString();
                    String dst = Operator.Move.getDest(quad).getRegister().toString();
                    graph.addMovePair(src, dst);
                }
            }
            // removing the definition:
            List.RegisterOperand defSet = quad.getDefinedRegisters();
            for(int i = 0; i < defSet.size(); i++) {
                graph.newNode(defSet.getRegisterOperand(i).getRegister().toString());
                liveVariables.remove(defSet.getRegisterOperand(i).getRegister().toString());
            }
            // adding all the uses:
            List.RegisterOperand usedSet = quad.getUsedRegisters();
            for(int i = 0; i < usedSet.size(); i++) {
                graph.newNode(usedSet.getRegisterOperand(i).getRegister().toString());
                liveVariables.add(usedSet.getRegisterOperand(i).getRegister().toString());
            }
            java.util.Iterator iterator = liveVariables.iterator();
            String[] regList = new String[liveVariables.size()];
            int i = 0;
            while(iterator.hasNext()) {
                regList[i] = (String)iterator.next();
                i++;
            }
            for(i = 0; i < (regList.length - 1); i++)
                for(int j = i + 1; j < regList.length; j++) {
                    String r1 = regList[i];
                    String r2 = regList[j];
                    graph.addEdge(r1, r2);
                }
        }
    }


    /**
     * This method computes the set of interferences of each basic block.
     */
    private void buildIntraBlockInterferences(ControlFlowGraph cfg) {
        ListIterator.BasicBlock bbIterator = cfg.postOrderOnReverseGraphIterator();
        while(bbIterator.hasNext()) {
            BasicBlock bb = bbIterator.nextBasicBlock();
            buildInterferencesInBlock(bb);
        }
    }


    /**
     * This method color the interference graph and prints some statistics
     * related to the solution found.
     * @param numColors the number of colors that will be used in the coloring.
     * @param methodName the name of the method that gave origin to the code being
     * used in the creation of the control flow graph.
     * @param fileName the name of the file in which the interference graph will
     * be saved.
     */
    public void printStatistics(int numColors, String fileName, String methodName) {
        graph.colorGraph(numColors);
        graph.printStatistics(methodName, System.out);
    }


    /**
     * This method simply runs the coloring algorithm. It is useful to take
     * the running time of the coloring method. It does not generate any
     * output.
     * @param numColors the number of colors that will be used in the coloring.
     * @param methodName the name of the method that gave origin to the code being
     * used in the creation of the control flow graph.
     * @param fileName the name of the file in which the interference graph will
     * be saved.
     */
    public void runAlgorithm(int numColors, String fileName, String methodName) {
        try {
            graph.colorGraph(numColors);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    /**
     * This method color the interference graph and saves the solution of the
     * coloring in a file.
     * @param numColors the number of colors that will be used in the coloring.
     * @param methodName the name of the method that gave origin to the code being
     * used in the creation of the control flow graph.
     * @param fileName the name of the file in which the interference graph will
     * be saved.
     */
    public void saveToFile(int numColors, String fileName, String methodName) {
        try {
            graph.toGraphviz(fileName, methodName);
        } catch (IOException e) {
            System.err.println("Error: probably the directory results does not exist.\n");
        }
    }


    /**
     * This is the main method of this visitor class.
     */
    public void visitCFG(ControlFlowGraph cfg) {
        if(this.isSSA) {
            CFGSculptor sculp = new CFGSculptor(cfg);
            sculp.adaptCFG();
        }
        buildInterBlockInterferences(cfg);
        buildIntraBlockInterferences(cfg);
        String methodName = cfg.getMethod().getName().toString();
        int randNumber = (int) (java.lang.Math.random()*Integer.MAX_VALUE);
        String fileName = "results/" + this.className + '_' + methodName + '_' + randNumber + ".dot";
        if(toGraph) {
            this.saveToFile(this.numRegisters, fileName, methodName);
        } else {
//            this.runAlgorithm(this.numRegisters, fileName, methodName);
            this.printStatistics(this.numRegisters, fileName, methodName);
        }
    }
}
